	/*

	Copyright (C) 1998 Stefan Westerfeld
                       stefan@space.twc.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    */

#include <sys/types.h>
#include <sys/time.h>
#include <math.h>
#include <stdio.h>
#include <qlist.h>
#include <qdict.h>
#include <qstring.h>
#include <ctype.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include "synth_impl.h"
 
extern "C" {
#include "sound.h"
#include "utils.h"
}

#define Die(a) { fprintf(stderr,"ERROR: %s\n",a); exit(1); }

//****************** KSynthesizer *******************

void strip_spaces(char *buffer)
{
	int sp=0,dp=0;
	while(buffer[sp] != 0)
	{
		switch(buffer[sp])
		{
			case ' ':
			case '\t':
			case '\n':
				break;
			default:
				buffer[dp++] = buffer[sp];
		}
		sp++;
	}
	buffer[dp] = 0;
}

float *KSynthesizer_impl::AllocArg(QString name, int flags)
{
	float *pos = ArgNames[name];

	fprintf(stderr,"dbg: allocarg %s = %lx\n",name.data(),pos);
	if(!pos)
	{
		pos = new float;
		ArgNames.insert(name,pos);
		ArgNameList.append(new QString(name));
		ArgFlags.insert(name,new int(0));
	}
	if(name.toFloat() != 0)
	{
		*pos = name.toFloat();
		flags |= ARG_CONST;
	}
	int *aflags = ArgFlags.take(name);
	*aflags = *aflags | flags;
	ArgFlags.insert(name,aflags);
	return(pos);
}
/*
SynthModule *KSynthesizer_impl::CreateModule(QString name)
{
	if(name == "Synth_RC")		return(Synth_RC::Create());
	if(name == "Synth_STDIN")	return(Synth_STDIN::Create());
	if(name == "Synth_DATA")	return(Synth_DATA::Create());
	if(name == "Synth_PLAY")	return(Synth_PLAY::Create());
	if(name == "Synth_STDOUT")	return(Synth_STDOUT::Create());
	if(name == "Synth_XFADE")	return(Synth_XFADE::Create());
	if(name == "Synth_NIL")		return(Synth_NIL::Create());
	if(name == "Synth_MUL")		return(Synth_MUL::Create());
	if(name == "Synth_ADD")		return(Synth_ADD::Create());
	if(name == "Synth_NET_MIDI")		return(Synth_NET_MIDI::Create());
	if(name == "Synth_PSCALE")	return(Synth_PSCALE::Create());
	if(name == "Synth_DELAY")	return(Synth_DELAY::Create());
	if(name == "Synth_FREQUENCY")return(Synth_FREQUENCY::Create());
	if(name == "Synth_FM_SOURCE")return(Synth_FM_SOURCE::Create());
	if(name == "Synth_WAVE_SIN")return(Synth_WAVE_SIN::Create());
	if(name == "Synth_WAVE_TRI")return(Synth_WAVE_TRI::Create());
	if(name == "Synth_SEQUENCE")return(Synth_SEQUENCE::Create());

	if(name == "Interpreter")	return(S_Interpreter::Create());
	fprintf(stderr,"ERROR: can't create requested type %s\n",name.data());
	return(NULL);
}
*/

void KSynthesizer_impl::Execute_Create(char *buffer)
{
	char *name, *type, *arg;

	strip_spaces(buffer);
	name = strtok(buffer,"=;\n");
	type = strtok(NULL,"(");

	SynthModule *NewModule = (SynthModule *)MS_SynthModule->getModule(type);
	if(!NewModule)
	{
		fprintf(stderr,"moduletype %s undefined\n",type);
		return;
		Die("moduletype undefined");
	}
	NewModule->setName(name);
	NewModule->Synthesizer = this;

// alloc arguments and link modules accordingly

	QString argname,textargs;

	NewModule->in_count = 0; NewModule->in = NULL;
	while(arg = strtok(NULL,",)"))
	{
		QString argval(arg);
		argname = NewModule->getArgument(NewModule->in_count,SynthModule::arg_in);
		//printf("%s = %s\n",argname.data(),argval.data());
		if(!argval.isEmpty() && argval != ";")
		{
			if(*argval.data() != '"')
			{
				NewModule->in = (float **)realloc(NewModule->in,
								sizeof(float*)*(NewModule->in_count+1));
				NewModule->in[NewModule->in_count] = AllocArg(argval,ARG_READ);
				NewModule->in_count++;
			}
			else
			{
				textargs = argval.mid(1,argval.length()-2);
				textargs.detach();
			}
		}
	}

	NewModule->out_count = 0; NewModule->out = NULL;
	do
	{
		argname = NewModule->getArgument(NewModule->out_count,SynthModule::arg_out);
		if(!argname.isEmpty())
		{
			NewModule->out = (float **)realloc(NewModule->out,
							sizeof(float*)*(NewModule->out_count+1));
			NewModule->out[NewModule->out_count] =
				AllocArg(name+QString(".")+argname, ARG_WRITE);
			NewModule->out_count++;
		}
	}
	while(!argname.isEmpty());

	NewModule->Initialize(textargs);
	SynthModules = (SynthModule **)realloc(SynthModules,
									sizeof(SynthModule *)*(MCount+1));
	SynthModules[MCount++] = NewModule;
}

void KSynthesizer_impl::Execute_Data(char *buffer)
{
	char *name = strtok(buffer," ;\n");
	QString data(strtok(NULL,";\n"));
	int i;

	for(i=0;i<MCount;i++)
	{
		if(SynthModules[i]->getName() == name)
		{
			SynthModules[i]->ProcessData(data);
		}
	}
	//fprintf(stderr,"got data for %s: %s\n",name,data.data());
}

void KSynthesizer_impl::Execute_Clear()
{
	fprintf(stderr,"*** clearing everything ***\n");

	if(SynthModules)
	{
		int i;
		printf("clear: deleting %d SynthModules\n",MCount);
		for(i=0;i<MCount;i++)
		{
			SynthModules[i]->DeInitialize();
			delete SynthModules[i];
		}

		free(SynthModules);
		SynthModules = NULL;
	}

	MCount = 0;
	// free memory ?

	ArgNames.clear();
	ArgNameList.clear();
	ArgFlags.clear();
}

void KSynthesizer_impl::Execute(const char *line)
{
	char buffer[1024];

	strcpy(buffer,line);

	//fprintf(stderr,"executing line '%s'\n",buffer);

	if(strlen(buffer) && buffer[0] != '#')
	{
		QString command = strtok(buffer," \t\n");
		char *args = strtok(NULL,"\n");

		if(args) strcpy(buffer,args);

		if(command == "create") Execute_Create(buffer);
		if(command == "data")	Execute_Data(buffer);
		if(command == "clear")	Execute_Clear();
	}
}

KSynthesizer_impl::KSynthesizer_impl()
{
	printf("... KSynthesizer_impl created\n");
	SynthModules = NULL;
	MCount = 0;
	MS_SynthModule = (ModuleServer<SynthModule> *)SynthModule::get_MS();
}

KSynthesizer_impl::~KSynthesizer_impl()
{
	printf("... KSynthesizer_impl destroyed\n");
	// mainly here to make my gcc (gcc 2.7.2.3) compile with -O2
}

CORBA::Boolean KSynthesizer_impl::SyntaxCheck( char *&error )
{
	QString *a;

	printf(">> SyntaxCheck\n");
	for (a=ArgNameList.first();a;a=ArgNameList.next())
	{
		switch(*ArgFlags[*a])
		{
			case ARG_WRITE:
				fprintf(stderr,"result of calculations in %s is never used!\n",a->data());
				break;
			case ARG_READ:
				fprintf(stderr,"variable %s is used but never initialized\n",a->data());
				break;
			case ARG_READ | ARG_WRITE:
			case ARG_READ | ARG_CONST:
				break;
			default:
				fprintf(stderr,"possible strange initialization/usage of %s (internal flags: %d)\n",a->data(),*ArgFlags[*a]);
		}
		//printf("GOT %s flags %d\n",a->data(),*ArgFlags[*a]);
	}
	printf("<< SyntaxCheck\n");
	error = NULL;
}

void KSynthesizer_impl::Run(CORBA::Long cycles)
{
	SynthModule *m;
	int i;
	long n=0;
	//fprintf(stderr,"Using %ld modules for synthesis\n",MCount);
	for (n=0;n<cycles;n++)
		for(i=0;i<MCount;i++) SynthModules[i]->Calculate();
}


CORBA::Long KSynthesizer_impl::getModuleCount()
{
	return(MS_SynthModule->getModuleCount());
}

void collectPorts(SequenceTmpl<CORBA::String_var> &ports,SynthModule *m,SynthModule::ArgType argtype)
{
	int i = 0;
	QString argname;

	do
	{
		argname = m->getArgument(i,argtype);
		if(!argname.isEmpty())
		{
			i++;
			ports.length(i);
			//printf("%s/inarg[%d] = %s\n",name,i-1,argname.data());
			ports[i-1] = CORBA::string_dup(argname.data());
		}
	} while(!argname.isEmpty());
}

CORBA::Boolean KSynthesizer_impl::getModuleInfo( CORBA::Long nr, ModuleInfo*& info )
{
	char *name=MS_SynthModule->getModuleName(nr);
	if(name)
	{
		SynthModule *m = (SynthModule *)MS_SynthModule->getModule(name);

		if(m)
		{
			info = new ModuleInfo;
			info->name = CORBA::string_dup(name);

			// collect inports & outports
			collectPorts(info->inports,m,SynthModule::arg_in);
			collectPorts(info->outports,m,SynthModule::arg_out);
			return(true);
		}
	}
	return(false);
}
